#include <stdint.h>
#include <avr/io.h>
#include <util/delay.h>
#include "sd.h"

#define DESELECT() do { PORTB |= _BV(2); } while(0)
#define SELECT() do { PORTB &= ~_BV(2); } while(0)

//#define spi_writebyte(data) do { SPDR = data; while( !(SPSR & (1<<SPIF) ) ); } while(0)
inline uint8_t spi_writebyte(uint8_t data)
{

    SPDR = data;
    while (!(SPSR & (1 << SPIF))) ;

    return SPDR;
}

#define spi_readbyte() spi_writebyte(0xff)

uint8_t sd_command(uint8_t cmd, uint32_t arg, uint8_t crc);

uint8_t sd_idle_state(void)
{

    uint8_t i, retry, r1;

    SELECT();
    _delay_ms(1);
    for (i = 0; i < 20; i++)
        spi_writebyte(0xff);

    retry = 0;
    do {
        r1 = sd_command(0, 0, 0x95);
        retry++;
    } while (r1 != 1 && retry < 100);
    if (retry >= 100)
        return 1;
    retry = 0;
    do {
        r1 = sd_command(1, 0, 0x95);
    } while (r1 && retry < 100);
    if (retry >= 100)
        return 2;

    r1 = sd_command(59, 0, 0x95);
    r1 = sd_command(16, 512, 0x95);

    _delay_ms(1);
    SPCR &= ~(0x3);
    DESELECT();

    return 0;
}

inline uint8_t sd_command(uint8_t cmd, uint32_t arg, uint8_t crc)
{

    register uint8_t r1, retry = 0;
    DESELECT();
    spi_writebyte(0xff);
    spi_writebyte(0xff);
    spi_writebyte(0xff);
    SELECT();
    spi_writebyte(cmd | 0x40);
    spi_writebyte(0xff & (arg >> 24));
    spi_writebyte(0xff & (arg >> 16));
    spi_writebyte(0xff & (arg >> 8));
    spi_writebyte(0xff & arg);
    spi_writebyte(crc);

    do {
        r1 = spi_writebyte(0xff);
        retry++;
    } while (r1 == 0xff && retry < 100);

    return r1;
}

/*
uint8_t sd_read_sector( uint32_t addr, uint8_t *p ) {

    uint8_t r1=0;
    uint16_t i; 
    addr <<= 9;
    r1 = sd_command( 17, addr, 0x95 );     //读出RESPONSE     
//    { char buf[32]; sprintf( buf, "r1=%d", r1 ); uart_puts(buf); }
    while ( spi_writebyte(0xff) != 0xfe); //直到读取到了数据的开始头0XFE,才继续
    for( i = 0; i < 512; i++) {
        *p = spi_readbyte();
        p++;
    }
    spi_readbyte();
    spi_readbyte();
    DESELECT();
    return(r1);
}
*/
uint8_t sd_read_data(register uint8_t * dest,
                     uint32_t addr,
                     register uint16_t offset, register uint16_t count)
{

    uint8_t r1 = 0;
    register uint16_t n;

    n = 514 - offset - count;
    addr <<= 9;
    r1 = sd_command(17, addr, 0x95);    //读出RESPONSE     
    if (r1 != 0)
        return -1;
    while (spi_writebyte(0xff) != 0xfe) ;   //直到读取到了数据的开始头0XFE,才继续

    for (; offset > 0; offset--)
        spi_readbyte();         // skip leading bytes

    for (; count > 0; count--) {
        *dest = spi_readbyte();
        dest++;
    }

    for (; n > 0; n--) {
        spi_readbyte();         // skip trailing bytes and crc
    }

    DESELECT();
    return (r1);
}
